# NOTE: this file depends on 'test-includes.r', sourced in utilitiesR/tests/test-all.r
# source('test-includes.r')
context("development functions")

describe('my.package.skeleton', {
    # returns the basic list of files and dirs (not any special ones)
    getExpectedFiles <<- function(packageName, packageDir, build.dir) {
        expected.dirs <- c(packageDir,
                           file.path(packageDir, build.dir),
                           file.path(packageDir, packageName),
                           file.path(packageDir, packageName, 'R'),
                           file.path(packageDir, packageName, 'man'))
        expected.files <- c(file.path(packageDir, 'Makefile'),
                            file.path(packageDir, 'README.md'))
        return(list(dirs=expected.dirs, files=expected.files))
    }
    expectFilesDirsExist <<- function(obj) {
        for (f in obj$files)
            expect_that(f, is_file())

        for (d in obj$dirs)
            expect_that(d, is_directory())
    }

    test("exists and is a function", {
        expect_that('my.package.skeleton', is_exported('function'))
    })

    test("creates the appropriate files (default options)", {
        f  <- tempfile(pattern='package')
        pname <- basename(f)
        basedir <- dirname(f)
        my.package.skeleton(pname, path=basedir)

        pdir <- sprintf('%s-package', pname)
        expectFilesDirsExist(getExpectedFiles(pname, file.path(basedir, pdir), 'bin'))
        unlink(f, recursive=T)
    })

    test("creates the appropriate files (build.dir argument)", {
        f  <- tempfile(pattern='package')
        pname <- basename(f)
        basedir <- dirname(f)
        my.package.skeleton(pname, path=basedir, build.dir='build')

        pdir <- sprintf('%s-package', pname)
        expectFilesDirsExist(getExpectedFiles(pname, file.path(basedir, pdir), 'build'))
        unlink(f, recursive=T)
    })

    test("creates the appropriate files (hg/hg.commit argument)", {
        # skip if no HG installed
        if (any(nchar(Sys.which(c('hg.exe', 'hg'))))) {
            f  <- tempfile(pattern='package')
            pname <- basename(f)
            basedir <- dirname(f)
            pdir <- sprintf('%s-package', pname)
            my.package.skeleton(pname, path=basedir, hg=T)

            obj <- getExpectedFiles(pname, file.path(basedir, pdir), 'bin')
            obj$dirs <- c(obj$dirs, file.path(basedir, pdir, '.hg'))
            obj$files <- c(obj$files, file.path(basedir, pdir, '.precommit.sh'))
            # don't know how to test `hg commit`, won't bother.

            expectFilesDirsExist(obj)
            unlink(f, recursive=T)
        }
    })

    test("creates the appropriate files ('force' argument)", {
        f <- tempfile(pattern='package')
        pname <- basename(f)
        basedir <- dirname(f)
        pdir <- sprintf('%s-package', pname)

        # verify the files were made
        my.package.skeleton(pname, path=basedir)
        expectFilesDirsExist(getExpectedFiles(pname, file.path(basedir, pdir), 'bin'))

        # with force=F should throw an error
        expect_that(my.package.skeleton(pname, path=basedir, force=FALSE), throws_error('already exists'))

        # now use force=T, should have no errors.
        expect_that(my.package.skeleton(pname, path=basedir, force=TRUE), not(throws_error()))
        unlink(file.path(basedir, pname), recursive=T)
    })

    # TODO: test 'empty' and 'hg.commit' arguments.
    test("creates the appropriate files ('version' argument)", {
        f  <- tempfile(pattern='package')
        pname <- basename(f)
        basedir <- dirname(f)
        my.package.skeleton(pname, path=basedir, version='0.2')

        pdir <- sprintf('%s-package', pname)

        DESC <- readLines(file.path(basedir, pdir, pname, 'DESCRIPTION'))
        expect_that(DESC, matches('Version: 0.2', all=F))
        unlink(f, recursive=T)
    })
})

describe('reload', {
    test("exists and is a function", {
        expect_that('reload', is_exported('function'))
    })
    # TODO: unsure how to test
})

describe('muffle', {
    test("exists and is a function", {
        expect_that('muffle', is_exported('function'))
    })

    test("suppresses all output except errors", {
        expect_that(muffle(message('boo')), not(shows_message()))
        expect_that(muffle(warning('boo')), not(gives_warning()))
        expect_that(muffle(packageStartupMessage('boo')), not(shows_message()))
        expect_that(muffle(cat('foo')), not(prints_text('foo')))
        expect_that(muffle(print('foo')), not(prints_text('foo')))

        expect_that(muffle(stop('boo')), throws_error())
    })
})

describe('muffle.messages', {
    test("exists and is a function", {
        expect_that('muffle.messages', is_exported('function'))
    })

    test("suppresses output from `message`, `warning` and `packageStartupMessage`", {
        expect_that(muffle.messages(message('boo')), not(shows_message()))
        expect_that(muffle.messages(warning('boo')), not(gives_warning()))
        expect_that(muffle.messages(packageStartupMessage('boo')), not(shows_message()))
    })

    test("does not suppress any other output", {
        expect_that(muffle.messages(stop('boo')), throws_error())
        expect_that(muffle.messages(cat('foo')), prints_text('foo'))
        expect_that(muffle.messages(print('foo')), prints_text('foo'))
    })
})

describe('stringify', {
    test("exists and is a function", {
        expect_that('stringify', is_exported('function'))
    })

    test("acts like dput", {
        expect_that(stringify('a'), is_identical_to('"a"'))
        expect_that(stringify(1), is_identical_to('1'))
        expect_that(stringify(NULL), is_identical_to('NULL'))
        expect_that(stringify(1:3), is_identical_to('1:3'))
        expect_that(stringify(c(1,2,3)), is_identical_to('c(1, 2, 3)'))
        expect_that(stringify(list(a=2, b=3)),
                    is_identical_to('structure(list(a = 2, b = 3), .Names = c("a", "b"))'))
    })

    test("is reversible with evaluate", {
        x <- data.frame(a=runif(10), b=1:10, c=letters[1:10])
        y <- stringify(x)
        expect_that(evaluate(y), equals(x))
    })
})

describe('loader', {
    #fixtures/loader.rda has a=1, b=2, c=3
    test("exists and is a function", {
        expect_that('loader', is_exported('function'))
    })

    test("returns a list with the loaded objects, keyed to the object names", {
        out <- loader('fixtures/loader.rda')
        expect_that(out, is_a('list'))
        expect_that(names(out), equals(c('a', 'b', 'c')))
        expect_that(out$a, equals(1))
        expect_that(out$b, equals(2))
        expect_that(out$c, equals(3))
    })

    test("only returns the objects asked for if the `names` parameter is specified", {
        out <- loader('fixtures/loader.rda', c('a', 'foobar'))
        expect_that(out, is_a('list'))
        # foobar is not in out
        expect_that(names(out), equals('a'))
        expect_that(out$a, equals(1))
    })
})
