# NOTE: this file depends on 'test-includes.r', sourced in utilitiesR/tests/test-all.r
# source('test-includes.r')
context("String functions")

describe('trim', {
    test("exists and is a function", {
        expect_that('trim', is_exported('function'))
    })

    test("respects the 'side' argument", {
        inp <- '  \n  a \t\r   '
        expect_that(trim(inp, side='both'), is_identical_to('a'))
        expect_that(trim(inp, side='left'), is_identical_to('a \t\r   '))
        expect_that(trim(inp, side='right'), is_identical_to('  \n  a'))
    })

    test("respects the 'what' argument", {
        inp <- '\n  a \t\r'
        expect_that(trim(inp, what='\n'), is_identical_to('  a \t\r'))
        expect_that(trim(inp, what='[\r\n]'), is_identical_to('  a \t'))
    })

    test("respects the 'perl' argument", {
        inp <- '!!  a !..'
        expect_that(trim(inp, what='(\\.|!(?=!))', perl=T), is_identical_to('!  a !'))
    })

    # TODO: trimming NA?
})

describe('getLastChar', {
    test("exists and is a function", {
        expect_that('getLastChar', is_exported('function'))
    })

    test("returns the last character or '' if the input is ''", {
        expect_that(getLastChar('alkjasd'), is_identical_to('d'))
        expect_that(getLastChar(''), is_identical_to(''))
    })
})

describe('%+%', {
    test("exists and is a function", {
        expect_that('%+%', is_exported('function'))
    })

    test("concatenates strings, and works on vectors", {
        expect_that('a' %+% 'b', is_identical_to('ab'))
        expect_that('I do not like ' %+% c('green eggs and ham!', 'them, Sam I Am!'),
                    is_identical_to(c('I do not like green eggs and ham!',
                                      'I do not like them, Sam I Am!')))
        expect_that(c('foo', 'bar') %+% 'baz',
                    is_identical_to(c('foobaz', 'barbaz')))
        expect_that(c('a', 'b') %+% c('A', 'B'),
                    is_identical_to(c('aA', 'bB')))
    })
})

describe('collapse', {
    test("exists and is a function", {
        expect_that('collapse', is_exported('function'))
    })

    test("works like paste", {
        expect_that(collapse(letters), is_identical_to(paste(letters, collapse='')))
    })

    test("respects the 'sep' and 'collapse' arguments", {
        expect_that(collapse(letters[1:3], collapse='|'),
                    is_identical_to('a|b|c'))
        expect_that(collapse(letters[1:3], 1:3, sep='.'),
                    is_identical_to('a.1b.2c.3'))
        expect_that(collapse(letters[1:3], 1:3, sep='.', collapse='|'),
                    is_identical_to('a.1|b.2|c.3'))
    })
})

describe('regexp', {
    test("exists and is a function", {
        expect_that('regexp', is_exported('function'))
    })

    test("is vectorised over the text argument", {
        out <- regexp('^Hi (.+)$', c('Hi world', 'Hi foo', 'Hi bar'))
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(3))
        expect_that(ncol(out), equals(1))
        expect_that(out[, 1], equals(c('world', 'foo', 'bar')))
    })

    test("throws an error if there are no captured groups", {
        expect_that(regexp('.', letters), throws_error('No capturing groups detected'))
    })

    test("does not throw an error if there are capturing groups but no matches", {
        out <- regexp('(z)', c('a', 'z'))
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(2))
        expect_that(ncol(out), equals(1))
        expect_that(unname(out[, 1]), equals(c('', 'z')))
    })

    test("extracts multiple captured groups, mixture of named and unnamed, always returning a matrix", {
        out <- regexp('^(?<greeting>.+?)[, ]+(.+)$', c('Hello, world', 'Hi bar'))
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(2))
        expect_that(ncol(out), equals(2))
        expect_that(colnames(out), equals(c('greeting', '')))
        expect_that(out[, 1], equals(c("Hello", "Hi")))
        expect_that(out[, 2], equals(c("world", "bar")))

        # try on non-vectorized
        out <- regexp('Hello (?<name>.+)', 'Hello world')
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(1))
        expect_that(ncol(out), equals(1))
        expect_that(colnames(out), equals('name'))
        expect_that(unname(out[, 1]), equals('world'))
    })

    test("passes extra parameters to regexpr/gregexpr", {
        out <- regexp('^([a-z]+)', 'ASDF foo', ignore.case=T)
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(1))
        expect_that(ncol(out), equals(1))
        expect_that(unname(out[, 1]), equals("ASDF"))
    })

    test("does not fail when there are multiple capturing groups and only one test strin", {
        out <- regexp('(f)(oo)bar', 'foobar')
        expect_that(out, is_a("matrix"))
        expect_that(nrow(out), equals(1))
        expect_that(ncol(out), equals(2))
        expect_that(unname(out[1, 1]), equals('f'))
        expect_that(unname(out[1, 2]), equals('oo'))
    })
})

describe('gregexp', {
    test("exists and is a function", {
        expect_that('regexp', is_exported('function'))
    })

    test("returns a list of lists, matching multiple times within the same string", {
        out <- gregexp('([aeiou])', 'hello')
        expect_that(out, is_a('list'))
        expect_that(length(out), equals(1))
        expect_that(names(out), equals('hello'))
        expect_that(out$hello, is_a('list'))
        expect_that(length(out$hello), equals(1))
        expect_that(out$hello[[1]], equals(c('e', 'o')))
    })

    test("is vectorised over the text argument", {
        out <- gregexp('([aeiou])', c('hello', 'goodbye'))
        expect_that(out, is_a('list'))
        expect_that(length(out), equals(2))
        expect_that(names(out), equals(c('hello', 'goodbye')))

        expect_that(out$hello, is_a('list'))
        expect_that(length(out$hello), equals(1))
        expect_that(out$hello[[1]], equals(c('e', 'o')))

        expect_that(out$goodbye, is_a('list'))
        expect_that(length(out$goodbye), equals(1))
        expect_that(out$goodbye[[1]], equals(c('o', 'o', 'e')))
    })

    test("extracts multiple captured groups, naming the list of lists appropriately", {
        out <- gregexp('(?<vowel>[aeiou])|([^aeiou])', c('hello', 'goodbye'))
        expect_that(out, is_a('list'))
        expect_that(length(out), equals(2))
        expect_that(names(out), equals(c('hello', 'goodbye')))

        expect_that(out$hello, is_a('list'))
        expect_that(length(out$hello), equals(2))
        expect_that(names(out$hello), equals(c('vowel', '')))
        expect_that(out$hello$vowel, equals(c('e', 'o')))
        expect_that(out$hello[[2]], equals(c('h', 'l', 'l')))

        expect_that(out$goodbye, is_a('list'))
        expect_that(length(out$goodbye), equals(2))
        expect_that(names(out$goodbye), equals(c('vowel', '')))
        expect_that(out$goodbye$vowel, equals(c('o', 'o', 'e')))
        expect_that(out$goodbye[[2]], equals(c('g', 'd', 'b', 'y')))
    })

    test("does not name the top-level list if use.names=FALSE", {
        out <- gregexp('(string|long)', c('a really really long string', 'another string'), use.names=F)
        expect_that(out, is_a('list'))
        expect_that(names(out), is_null())

        expect_that(length(out), equals(2))

        expected <- list(list(c('long', 'string')), list('string'))
        names(expected[[1]]) <- ''
        names(expected[[2]]) <- ''
        expect_that(out, equals(expected))
    })

    test("throws an error if there are no captured groups", {
        expect_that(gregexp('.', 'asdf'), throws_error('No capturing groups detected'))
    })

    test("does not throw an error if there are capturing groups but no matches", {
        out <- gregexp('(z)', 'asdf')
        expect_that(out, is_a('list'))
        expect_that(length(out), equals(1))

        expect_that(out$asdf, is_a('list'))
        expect_that(length(out$asdf), equals(1))
        expect_that(length(out$asdf[[1]]), equals(0))
    })

    test("passes extra parameters to regexpr/gregexpr", {
        out <- gregexp('([aeiou])', 'HIi', ignore.case=T)

        expect_that(out, is_a('list'))
        expect_that(length(out), equals(1))
        expect_that(out$HIi, is_a('list'))
        expect_that(length(out$HIi), equals(1))
        expect_that(out$HIi[[1]], equals(c('I', 'i')))
    })
})

describe('regescape', {
    test("exists and is a function", {
        expect_that('regescape', is_exported('function'))
    })

    test("is vectorised", {
        expect_that(regescape(c('a', 'b', 'c')),
                    equals(c('a', 'b', 'c')))
        expect_that(regescape(c('a', 'b?', '^c[')),
                    equals(c('a', 'b\\?', '\\^c\\[')))
    })

    test("escapes special characters .*+?^$[](){}|- and blackslash", {
        expect_that(regescape('.*+?^$[](){}|-\\'),
                    is_identical_to('\\.\\*\\+\\?\\^\\$\\[\\]\\(\\)\\{\\}\\|\\-\\\\'))
    })

    test("respects the `escape` argument.", {
        expect_that(regescape(c('&lt; #ffaabb !?.[]', 'a#.&'), '&#'),
                    equals(c('\\&lt; \\#ffaabb !?.[]', 'a\\#.\\&')))
        expect_that(regescape('.?!@', ''),
                    is_identical_to('.?!@'))
    })
})
