# NOTE: this file depends on 'test-includes.r', sourced in utilitiesR/tests/test-all.r
# source('test-includes.r')
context("Misc. utilities")


describe('find.in', {
    test("exists and is a function", {
        expect_that('find.in', is_exported('function'))
    })
    test("if value=T and inv=F, returns x[x %in% y]", {
         x <- runif(10)
         y <- sample(x, size=6)
         expect_that(find.in(x, y, inv=F, value=T), equals(x[x %in% y]))
    })
    test("if value=T and inv=T, returns x[!(x %in% y)]", {
         x <- runif(10)
         y <- sample(x, size=6)
         expect_that(find.in(x, y, inv=T, value=T), equals(x[!(x %in% y)]))
    })
    test("if value=F and inv=F, returns x %in% y]", {
         x <- runif(10)
         y <- sample(x, size=6)
         expect_that(find.in(x, y, inv=F, value=F), equals(x %in% y))
    })
    test("if value=F and inv=T, returns x[!(x %in% y)]", {
         x <- runif(10)
         y <- sample(x, size=6)
         expect_that(find.in(x, y, inv=T, value=F), equals(!(x %in% y)))
    })
})

#' Opposite to \code{which}.
#'
#' @param indices the indices that should be TRUE
#' @param length.out the length of the output vector. if NULL, `max(indices)` is used.
#' @param use.names whether to use the names of `indices` if there are any.
#' @author user Nick Sabbe on StackOverflow: http://stackoverflow.com/a/7661128/913184
#' @return a logical vector of length `length.out` that is TRUE at the indices
#'         in `indices`.
#' @examples
#' idx <- c(1, 3)
#' invwhich(idx)     # c(TRUE, FALSE, TRUE)
#' invwhich(idx, 5)  # c(TRUE, FALSE, TRUE, FALSE, FALSE)
#'
#' ## example with use.names
#' names(idx) <- c('one', 'three')
#' inwhich(idx, 4)
#' @family membership
#' @export
describe('invwhich', {
    test("exists and is a function", {
        expect_that('invwhich', is_exported('function'))
    })

    test("returns a logical vector of length `length.out`", {
         ix <- which(runif(10) > .5)
         y <- invwhich(ix, 10)
         expect_that(is.vector(y), is_true())
         expect_that(mode(y), is_identical_to("logical"))
         expect_that(length(y), equals(10))
    })

    test("returns a logical vector of length `max(indices)` if length.out is not provided", {
         ix <- which(runif(10) > .5)
         y <- invwhich(ix)
         expect_that(is.vector(y), is_true())
         expect_that(mode(y), is_identical_to("logical"))
         expect_that(length(y), equals(max(ix)))
    })

    test("satisfies invwhich(which(x), length(x)) == as.logical(x)", {
         x <- runif(10) > .5
         ix <- which(x)
         y <- invwhich(ix, length(x))
         expect_that(y, equals(x))
    })

    test("sets the `names` attribute if use.names=T", {
         ix <- c(1, 3)
         names(ix) <- c('one', 'three')
         y <- invwhich(ix, length.out=4)
         expect_that(y, is_equivalent_to(c(T, F, T, F)))
         expect_that(names(y), equals(c('one', NA, 'three', NA)))
    })
})

describe('%notin%', {
    test("exists and is a function", {
        expect_that('%notin%', is_exported('function'))
    })

    test("returns a logical vector of length x", {
        x <- 1:5
        out <- x %notin% c(1, 3, 5)
        expect_that(is.vector(out), is_true())
        expect_that(length(out), equals(length(x)))
        expect_that(mode(out), equals('logical'))
    })

    test("is equivalent to !(x %in% table)", {
         x <- 1:1
         0
         table <- sample(10, size=5, replace=T)
         expect_that(x %notin% table, equals(!(x %in% table)))
    }, times=10)
})

describe('is.empty', {
    test("exists and is a function", {
        expect_that('is.empty', is_exported('function'))
    })
    
    test("returns TRUE iff the input is null, NA or length 0", {
        expect_that(is.empty(NA), is_true())
        expect_that(is.empty(NULL), is_true())
        expect_that(is.empty(character()), is_true())
        expect_that(is.empty(TRUE), is_false())
        expect_that(is.empty(FALSE), is_false())
        expect_that(is.empty(''), is_false())
    })
})

describe('make2Drot', {
    test("exists and is a function", {
        expect_that('make2Drot', is_exported('function'))
    })

    test("makes a 2D rotation matrix...", {
         # just pick out a few cases?
         expect_that(make2Drot(pi/2), equals(array(c(0, 1, -1, 0), dim=c(2, 2))))
         expect_that(make2Drot(pi), equals(array(c(-1, 0, 0, -1), dim=c(2, 2))))
         expect_that(make2Drot(3*pi/2), equals(array(c(0, -1, 1, 0), dim=c(2, 2))))
    })
})

describe('euclidDist', {
    test("exists and is a function", {
        expect_that('euclidDist', is_exported('function'))
    })

    # nxm matrix, n points in R^m, distance between rows.
    test("calculates euclidean distance of matrices of various sizes", {
         inMat <- matrix(runif(8), ncol=2) # 4 points
         out   <- euclidDist(inMat)
         expect_that(is.vector(out), is_true())
         expect_that(length(out), equals(3))
         
         expect_that(out[1], equals(sqrt(sum((inMat[1, ] - inMat[2, ])^2))))
         expect_that(out[2], equals(sqrt(sum((inMat[2, ] - inMat[3, ])^2))))
         expect_that(out[3], equals(sqrt(sum((inMat[3, ] - inMat[4, ])^2))))

         inMat <- matrix(runif(6), ncol=3) # 2 points in R3
         out   <- euclidDist(inMat)
         expect_that(is.vector(out), is_true())
         expect_that(length(out), equals(1))
         
         expect_that(out[1], equals(sqrt(sum((inMat[1, ] - inMat[2, ])^2))))
    })

    test("respects `cols` argument", {
         inMat <- matrix(runif(4), ncol=2) # 2 points
         out   <- euclidDist(inMat, cols=T)
         expect_that(is.vector(out), is_true())
         expect_that(length(out), equals(1))
         
         expect_that(out[1], equals(sqrt(sum((inMat[, 1] - inMat[, 2])^2))))
    })

    test("respects `circular` argument", {
         inMat <- matrix(runif(6), ncol=2) # 3 points
         out   <- euclidDist(inMat, circular=T)
         expect_that(is.vector(out), is_true())
         expect_that(length(out), equals(3))
         
         expect_that(out[1], equals(sqrt(sum((inMat[1, ] - inMat[2, ])^2))))
         expect_that(out[2], equals(sqrt(sum((inMat[2, ] - inMat[3, ])^2))))
         expect_that(out[3], equals(sqrt(sum((inMat[3, ] - inMat[1, ])^2))))
    })
})

# BIG FAILURES: preserveStart is broken.
describe('sortPointsRectangle', {
    # makes a rotated rectangle
    makeRect <<- function (th) {
        rec <- matrix(c(1, 1,
                        5, 1,
                        5, 3,
                        1, 3), ncol=2, byrow=T) 
        colnames(rec) <- c('X', 'Y')
        return(t(make2Drot(th) %*% t(rec)))
    }

    test("exists and is a function", {
        expect_that('sortPointsRectangle', is_exported('function'))
    })

    test("sorts points for exact rectangles", {
        # NOTE: for corner cases the output is NOT CONSISTENT but depends on
        # which of the points equidistant to 5*pi/4 occurs first in pts.
        for (i in 1:360) {
            if (i %in% c(45, 135, 225, 315)) {
                next
            }
            th <- i / 180 * pi
            inp <- makeRect(th)
            IDX <- sample.int(4) # shake up so the points are out of sequence
            inp <- inp[IDX, ]
            # index.return: order should be preserved
            out <- sortPointsRectangle(inp, preserveStart=T, index.return=T)
            expect_that(out, is_a('list'))
            expect_that(names(out), equals(c('x', 'ix')))
            expect_that(out$x, is_equivalent_to(inp[out$ix, ]))

            # check that with index.return the results agree
            tmp <- sortPointsRectangle(inp, preserveStart=T)
            expect_that(out$x, equals(tmp))

            # check labels for a few easy cases...
            rn <- NULL
            if (i > 315 || i < 45) {
               rn <- c('bottom-left', 'bottom-right', 'top-right', 'top-left')
            } else if (i > 45 && i < 135) {
               rn <- c('bottom-right', 'top-right', 'top-left', 'bottom-left')
            } else if (i > 135 && i < 225) {
               rn <- c('top-right', 'top-left', 'bottom-left', 'bottom-right')
            } else if (i > 225 && i < 315) {
               rn <- c('top-left', 'bottom-left', 'bottom-right', 'top-right')
            }
            # inp0[IDX] = inp
            # inp[out$ix] = out
            # hence do order(indices) to reverse from out back to inp.
            expect_that(rownames(out$x)[order(out$ix)][order(IDX)], equals(rn))
            if (!isTRUE(all.equal(rownames(out$x)[order(out$ix)][order(IDX)], rn))) {
                cat('degrees:', i, '\n')
                cat('IDX:', IDX, '\n')
            }
        }
    })

    test("respects colnames X and Y, if provided", {
         inp <- makeRect(43 / 180 * pi)
         colnames(inp) <- c('X', 'Y')
         out <- sortPointsRectangle(inp)
         expect_that(inp, is_equivalent_to(out))
         expect_that(rownames(out),
                     equals(c('bottom-left', 'bottom-right', 'top-right', 'top-left')))

         # switch X & Y, should get the same result.
         inp <- inp[, c('Y', 'X')]
         out2 <- sortPointsRectangle(inp)
         expect_that(out2, equals(out))
    })

    test("returns the requested order, taking into account anticlockwise and preserveStart", {
         rec <- jitter(makeRect(70)) # rec[4, ] is bottom-left
         out <- sortPointsRectangle(rec, anticlockwise=F)
         expect_that(out, is_equivalent_to(rec[c(4, 3, 2, 1), ]))
         expect_that(rownames(out),
                     equals(c('bottom-left', 'top-left', 'top-right', 'bottom-right')))

         rec <- jitter(makeRect(70))
         out <- sortPointsRectangle(rec, anticlockwise=F, preserveStart=T)
         expect_that(out, is_equivalent_to(rec[c(1, 4, 3, 2), ]))
         expect_that(rownames(out),
                     equals(c('bottom-right', 'bottom-left', 'top-left', 'top-right')))
    })

    test("translates points relative to the bottom-left if shift=T", {
         # try with & without preserveStart

         rec <- jitter(makeRect(70 * pi / 180)) # bottomLeft is rec[4, ]
         out <- sortPointsRectangle(rec, shift=T, preserveStart=T)
         # first corner is bottom-right.
         expect_that(rownames(out),
                     equals(c('bottom-right', 'top-right', 'top-left', 'bottom-left')))
         expect_that(out, is_equivalent_to(t(t(rec) - rec[4, ])))

         # try with shift && !preserveStart
         out <- sortPointsRectangle(rec, shift=T, preserveStart=F)
         expect_that(rownames(out),
                     equals(c('bottom-left', 'bottom-right', 'top-right', 'top-left')))
         expect_that(out, is_equivalent_to(t(t(rec) - rec[4, ])[c(4, 1, 2, 3), ]))
         
    })

    test("returns a list with elements ix (indices) and x (points) if index.return=T", {
         rec <- makeRect(runif(1) * 2 * pi)
         out <- sortPointsRectangle(rec, index.return=F)
         out2 <- sortPointsRectangle(rec, index.return=T)

         expect_that(out2, is_a("list"))
         expect_that(names(out2), equals(c('x', 'ix')))
         expect_that(out2$x, equals(out))
         expect_that(rec[out2$ix, ], is_equivalent_to(out))
    })
    # TODO: tests with more oddly-shaped quads.
    # TODO: tests with points in non-sequential order.
    # TODO: deprecate or implement `verbose`.
})

describe('getUTMZone', {
    test("exists and is a function", {
        expect_that('getUTMZone', is_exported('function'))
    })

    test("divides the earth in 6degree zones, [-180, 174), ..., [174, 180)", {
        expect_that(getUTMZone(seq(-180, 179.5, by=0.5)),
                    equals(rep(1:60, each=12)))
        expect_that(getUTMZone(seq(0, 359.5, by=0.5)),
                    equals(rep(c(31:60, 1:30), each=12)))
    })
})

describe('GPS2UTM/UTM2GPS', {
    test("are exported functions in utilitiesR", {
        expect_that('GPS2UTM', is_exported('function'))
        expect_that('UTM2GPS', is_exported('function'))
    })
})

has.rgdal <<- suppressPackageStartupMessages(require(rgdal))
has.proj <<- !is.null(getProgram('proj'))
if (!has.rgdal && !has.proj) {
    warning("Cannot test GPS2UTM or UTM2GPS as you do not have the rgdal R package or proj executable")
} else {
    describe('GPS2UTM', {
        # input: n x 2 matrix, LON then LAT
        test("takes in a nx2 matrix and returns an nx2 matrix with 'zone' and 'hemisphere' attributes", {
            LL <- matrix(c(runif(5, min=-180, max=180), runif(5, min=-90, max=90)),
                         ncol=2)
            out <- GPS2UTM(LL)
            expect_that(out, is_a("matrix"))
            expect_that(dim(out), equals(c(5, 2)))
            expect_that(colnames(out), equals(c('X', 'Y')))
            expect_that(attr(out, 'zone'), not(is_null()))
            expect_that(is.vector(attr(out, 'zone')), is_true())
            expect_that(length(attr(out, 'zone')), equals(5))
            expect_that(attr(out, 'hemisphere'), not(is_null()))
            expect_that(length(attr(out, 'hemisphere')), equals(5))
            expect_that(is.vector(attr(out, 'hemisphere')), is_true())
        })

        test("respects the 'using' parameter, and outputs agree", {
            out.rgdal <- NULL
            out.proj <- NULL
            LL <- matrix(c(runif(5, min=-180, max=180), runif(5, min=-90, max=90)),
                         ncol=2)
            if (has.rgdal) {
                # cause proj option to fail. If this succeeds it *must* have used rgdal.
                f <- overrideIn(GPS2UTM, getProgram=function(...) NULL)
                expect_that(out.rgdal <- f(LL, using='rgdal'), not(throws_error()))
            }

            if (has.proj) {
                # cause rgdal option to fail. If this succeeds it *must* have used proj
                f <- overrideIn(GPS2UTM, require=function(...) FALSE)
                expect_that(out.proj <- f(LL, using='proj'), not(throws_error()))
            }
            
            if (has.rgdal && has.proj) {
                expect_that(out.proj, equals(out.rgdal))
            }
        })

        # TODO (need some ground truth but how to get it other than from R,
        # defeating the purpose of the whole exercise?)
        #LL <- as.matrix(expand.grid(-180:180, -90:90))
        #write.table(LL, 'fixtures/latlong.csv', col.names=F, row.names=F)
        #LL <- as.matrix(read.table('fixtures/latlong.csv'))
        #UTMs <- as.matrix(read.table('fixtures/utms.csv'))
        # UPTO
        xtest("gives UTM coordinates relative to zone provided, allowed to be vectorised", {
             # only need one latitude across all longitudes to test
             idx <- LL[, 2] == -80
             out <- GPS2UTM(LL[idx, ], zone=getUTMZone(LL[idx, 1]), hemisphere='south')
             expect_that(out, is_equivalent_to(UTMs[idx, ]))

             # TODO: now force the wrong zone
        })

        xtest("gives UTM coordinates relative to the closest zone for each coord if zone=NULL", {
             # only need one latitude across all longitudes to test
             idx <- LL[, 2] == -80
             out <- GPS2UTM(LL[idx, ], zone=NULL, hemisphere='south')
             expect_that(out, is_equivalent_to(UTMs[idx, ]))
        })

        test("gives UTM coordinates relative to the zone of the mean longitude if zone is missing", {
             # pick some that span across zones.
        })

        test("gives UTM coordinates relative to hemisphere provided, allowed to be vectorised", {
        })

        test("gives UTM coordinates relative to the hemisphere for each coord if hemisphere=NULL", {
        })

        test("gives UTM coordinates relative to the hemisphere of the mean latitude if hemisphere is missing", {
        })

        test("throws an error if rgdal/proj is not installed (whichever relevant)", {
            LL <- matrix(c(runif(5, min=-180, max=180), runif(5, min=-90, max=90)),
                         ncol=2)
            # an odd way to override require
            f <- overrideIn(GPS2UTM, require=function(...) FALSE)
            expect_that(f(LL, using='rgdal'),
                        throws_error('You need to install the rgdal package'))

            f <- overrideIn(GPS2UTM, getProgram=function(...) NULL)
            expect_that(f(LL, using='proj'),
                        throws_error('You need to install proj'))
        })
    })

    describe('UTM2GPS', {
        # input: n x 2 matrix, easting then northing

        test("takes in a nx2 matrix and converts to GPS (nx2 matrix)", {
            UTM <- matrix(c(runif(5) * 5000, runif(5) * 5000),
                          ncol=2)
            out <- UTM2GPS(UTM, zone=56, hemisphere='south')
            expect_that(out, is_a("matrix"))
            expect_that(dim(out), equals(c(5, 2)))
            expect_that(colnames(out), equals(c('lon', 'lat')))
        })

        test("goes backwards if `inverse=TRUE`", {
        })

        test("respects the 'using' parameter", {
        })

        test("goes backwards if `inverse=TRUE`", {
        })

        test("uses `zone` if provided (UTM to GPS) or calculates it (GPS to UTM)", {
        })

        test("uses hemisphere if provided (UTM to GPS) or calculates it (GPS to UTM)", {
        })
    })
}

describe('formatGPS', {
    test("exists and is a function", {
        expect_that('formatGPS', is_exported('function'))
    })

    test("takes in a matrix, returns string in format %.10fE, %.10fS", {
        inp <- matrix(runif(10), ncol=2)
        expect_that(formatGPS(inp, verbose=F),
                    is_identical_to(paste(sprintf('%.10fE, %.10fS',
                                                  inp[, 1],
                                                  -inp[, 2]),
                                          collapse='\n')))
    })

    test("is verbose if requested", {
        inp <- matrix(runif(10), ncol=2)
        expect_that(formatGPS(inp, verbose=F), not(shows_message()))
        # the separating '\n' appears to be gobbled up...
        expect_that(invisible(formatGPS(inp, verbose=T)), shows_message('^([0-9.]+E, -[0-9.]+S\n){5}$'))
    })
})

# TODO: probably put this in ui.r (user interaction)
describe('verifyInput', {
    test("exists and is a function", {
        expect_that('verifyInput', is_exported('function'))
    })
})
